/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"

#include <memory>

#include "common/common.h"
#include "common/log.h"
#include "defs.h"
#include "errors.h"
#include "parser/ast.h"
#include "system/sm_meta.h"

void Analyze::check_having_conds(std::vector<TabCol> &query_cols, std::vector<TabCol> &group_cols,
                                 std::vector<Condition> &having_conds,
                                 std::vector<ColMeta> &all_cols) {
  for (auto &col : query_cols) {
    if (col.col_name == "*" && col.aggregation_type == AGG_COUNT_STAR) {
      continue;
    }

    auto &col_meta = *std::find_if(all_cols.begin(), all_cols.end(), [&col](ColMeta &col_meta) {
      return col_meta.tab_name == col.tab_name && col_meta.name == col.col_name;
    });

    if (col_meta.type == TYPE_STRING && col.aggregation_type == AGG_SUM) {
      throw InternalError("string type can not sum.");
    }

    // if (col.aggregation_type != AGG_NO && col.aggregation_type != AGG_COUNT &&
    //     col_meta.type == TYPE_STRING) {
    //   throw InternalError("string type can not sum, min or max.");
    // }
  }

  // Get raw values in where clause
  for (auto &cond : having_conds) {
    if (cond.lhs_col.aggregation_type == AGG_COUNT_STAR) {
      continue;
    }

    // Infer table name from column name
    cond.lhs_col = check_column(all_cols, cond.lhs_col);
    auto &lhs_col_meta = *std::find_if(all_cols.begin(), all_cols.end(), [&cond](ColMeta &col) {
      return cond.lhs_col.tab_name == col.tab_name && cond.lhs_col.col_name == col.name;
    });

    // 如果没有having的列没有聚合，也要报错。
    if (cond.lhs_col.aggregation_type == AGG_NO &&
        group_cols.end() ==
            std::find_if(group_cols.begin(), group_cols.end(), [&cond](TabCol &tab_col) {
              return cond.lhs_col.tab_name == tab_col.tab_name &&
                     cond.lhs_col.col_name == tab_col.col_name;
            })) {
      throw InternalError("having col should be in the group cols");
    }

    if (lhs_col_meta.type == TYPE_STRING && cond.lhs_col.aggregation_type == AGG_SUM) {
      throw InternalError("string type can not sum.");
    }
    // if (cond.lhs_col.aggregation_type != AGG_NO && cond.lhs_col.aggregation_type != AGG_COUNT &&
    //     lhs_col_meta.type == TYPE_STRING) {
    //   throw InternalError("string type can not sum, min or max.");
    // }

    if (!cond.is_rhs_val) {
      throw InternalError("having condition right should be value.");
    }

    TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
    auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
    ColType lhs_type = lhs_col->type;
    ColType rhs_type;

    cond.rhs_val.convert_to_type(lhs_col->type);
    cond.rhs_val.init_raw(lhs_col->len);
    rhs_type = cond.rhs_val.type;

    if (!check_type(lhs_type, rhs_type)) {
      throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
    }
  }
}

void Analyze::check_group_list(std::vector<TabCol> &query_cols, std::vector<TabCol> &group_cols,
                               std::vector<ColMeta> &all_cols) {
  if (group_cols.empty()) {  // 如果没有group by。那么select要么全聚合，要么没有聚合。
    bool has_aggregation = false;
    bool has_common = false;

    for (auto &col : query_cols) {
      if (col.aggregation_type != AGG_NO) {
        has_aggregation = true;
      } else {
        has_common = true;
      }
    }

    if (has_aggregation && has_common) {
      throw InternalError(
          "select column should not have both aggregation function and table column without group "
          "by");
    }

    return;
  }

  // group_cols去重 -> 防止出现 group by a,b,a 这种情况。
  for (auto iter = group_cols.begin(); iter != group_cols.end(); ++iter) {
    if (iter->aggregation_type != AGG_NO) {
      throw InternalError("group list should not have aggregation function.");
    }

    auto i = std::find_if(iter + 1, group_cols.end(), [iter](TabCol &col) {
      return iter->tab_name == col.tab_name && iter->col_name == col.col_name;
    });

    if (i != group_cols.end()) {
      group_cols.erase(i);
    }
  }

  // 检查group列是否歧义或者不在应有列中。
  for (auto &group_col : group_cols) {
    group_col = check_column(all_cols, group_col);
  }

  // 要么是聚合要么在group列中。
  for (auto &query_col : query_cols) {
    if (query_col.aggregation_type != AGG_NO) {
      continue;
    }

    auto iter = std::find_if(group_cols.begin(), group_cols.end(), [&query_col](TabCol &col) {
      return query_col.tab_name == col.tab_name && query_col.col_name == col.col_name;
    });

    if (iter == group_cols.end()) {
      throw InternalError("query col should be in the group by cols.");
    }
  }
}

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query
 */
std::shared_ptr<Query> Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse) {
  std::shared_ptr<Query> query = std::make_shared<Query>();
  if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse)) {
    /** TODO: 检查表是否存在 */

    for (auto &tab_name : x->tabs) {
      if (!sm_manager_->db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
      }
    }

    // 处理表名
    query->tables = std::move(x->tabs);
    // 处理target list，再target list中添加上表名，例如 a.id
    for (auto &sv_sel_col : x->cols) {
      TabCol sel_col = {sv_sel_col->tab_name, sv_sel_col->col_name, sv_sel_col->alias,
                        sv_sel_col->aggregation_type};

      query->cols.push_back(sel_col);
    }

    // 处理group列。
    if (x->has_group) {
      for (auto &sv_group_col : x->group->cols) {
        TabCol group_col = {sv_group_col->tab_name, sv_group_col->col_name, sv_group_col->alias,
                            sv_group_col->aggregation_type};
        query->group_cols.push_back(group_col);
      }
    }

    std::vector<ColMeta> all_cols;
    get_all_cols(query->tables, all_cols);

    if (query->cols.empty()) {
      throw InternalError("select column could not be empty.");
    }  // 只有一行，而且为 *，没有聚合。
    else if (query->cols.size() == 1 && query->cols.front().col_name == "*" &&
             query->cols.front().aggregation_type == AGG_NO) {
      // select all columns
      query->cols.clear();
      for (auto &col : all_cols) {
        TabCol sel_col = {col.tab_name, col.name, "",
                          AGG_NO};  // select * from，不需要聚合，更不需要别名alias.
        query->cols.push_back(sel_col);
      }
    } else {
      // infer table name from column name
      for (auto &sel_col : query->cols) {
        sel_col = check_column(all_cols, sel_col);  // 列元数据校验, 聚合可以通过检验。
      }
    }

    if (x->has_sort) {
      TabCol sort_col;
      for (auto &col : all_cols) {
        if (col.name.compare(x->order->cols->col_name) == 0) {
          sort_col = {col.tab_name, col.name, "", AGG_NO};
        }
      }

      query->sort_col = check_column(all_cols, sort_col);
    }
    // 检查select列和group列 ：
    // 1.select列要么在group列中，要么就是列的聚合.
    // 2.select如果没有group by，要么全都是聚合，要么全都不是聚合。
    // 3.group列需要合法，应该在all_cols中。
    check_group_list(query->cols, query->group_cols, all_cols);

    // 处理where条件
    get_clause(x->conds, query->conds, true);
    check_clause(query->tables, query->conds);

    // 还需要处理having条件。
    if (x->has_group) {
      get_clause(x->group->having_conds, query->having_conds, false);
    }

    // 检查select的聚合，having的聚合是否符合要求 ： SUM不能对string使用。
    check_having_conds(query->cols, query->group_cols, query->having_conds, all_cols);
  } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
    /** TODO: */
    for (auto &set_clause : x->set_clauses) {
      SetClause sc{TabCol{x->tab_name, set_clause->col_name, "", AGG_NO},
                   convert_sv_value(set_clause->val)};
      query->set_clauses.emplace_back(sc);
    }

    get_clause(x->conds, query->conds);
    check_clause(std::vector<std::string>{x->tab_name}, query->conds);
  } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
    // 处理where条件
    get_clause(x->conds, query->conds);
    check_clause({x->tab_name}, query->conds);
  } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
    // 处理insert 的values值
    for (auto &sv_val : x->vals) {
      query->values.push_back(convert_sv_value(sv_val));
    }
  } else {
    // do nothing
  }
  query->parse = std::move(parse);
  return query;
}

TabCol Analyze::check_column(const std::vector<ColMeta> &all_cols, TabCol target) {
  if (target.col_name == "*" && target.aggregation_type == AGG_COUNT_STAR) {
    return target;
  }

  if (target.tab_name.empty()) {
    // Table name not specified, infer table name from column name
    std::string tab_name;
    for (auto &col : all_cols) {
      if (col.name == target.col_name) {
        if (!tab_name.empty()) {
          throw AmbiguousColumnError(target.col_name);
        }
        tab_name = col.tab_name;
      }
    }
    if (tab_name.empty()) {
      throw ColumnNotFoundError(target.col_name);
    }
    target.tab_name = tab_name;
  } else {
    /** TODO: Make sure target column exists */
    bool col_found = false;
    for (auto &col : all_cols) {
      if (col.tab_name == target.tab_name && col.name == target.col_name) {
        col_found = true;
      }
    }
    if (col_found == false) {
      throw ColumnNotFoundError(target.col_name);
    }
  }
  return target;
}

void Analyze::get_all_cols(const std::vector<std::string> &tab_names,
                           std::vector<ColMeta> &all_cols) {
  for (auto &sel_tab_name : tab_names) {
    // 这里db_不能写成get_db(), 注意要传指针
    const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
    all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
  }
}

void Analyze::get_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds,
                         std::vector<Condition> &conds, bool is_where /* = true */) {
  conds.clear();
  for (auto &expr : sv_conds) {
    if (is_where && expr->lhs->aggregation_type != AGG_NO) {
      throw InternalError("where clause must not have aggregation function.");
    }

    Condition cond;
    // 别名不用copy，反正在条件部分的别名也没啥意义，但是聚合类型需要copy，
    // 虽然where子句用不到，但是having子句会用到。
    cond.lhs_col.tab_name = expr->lhs->tab_name;
    cond.lhs_col.col_name = expr->lhs->col_name;
    cond.lhs_col.aggregation_type = expr->lhs->aggregation_type;

    cond.op = convert_sv_comp_op(expr->op);

    if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
      cond.is_rhs_val = true;
      cond.is_rhs_column = false;
      cond.is_sub_query = false;
      cond.is_value_list = false;

      cond.rhs_val = convert_sv_value(rhs_val);
    } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
      cond.is_rhs_val = false;
      cond.is_rhs_column = true;
      cond.is_sub_query = false;
      cond.is_value_list = false;

      cond.rhs_col.tab_name = rhs_col->tab_name;
      cond.rhs_col.col_name = rhs_col->col_name;

    } else if (expr->rhs->is_sub_query) {
      cond.is_rhs_val = false;
      cond.is_rhs_column = false;
      cond.is_sub_query = true;
      cond.is_value_list = false;

      auto parse_tree = do_analyze(expr->rhs->sub_query);
      if (nullptr == std::dynamic_pointer_cast<ast::SelectStmt>(parse_tree->parse)) {
        throw InternalError("sub query is not a select sentence.");
      }

      cond.analyzed_sub_query = parse_tree;
    } else if (expr->rhs->is_value_list) {
      cond.is_rhs_val = false;
      cond.is_rhs_column = false;
      cond.is_sub_query = false;
      cond.is_value_list = true;
      cond.value_list.reserve(expr->rhs->value_list.size());

      for (auto &val : expr->rhs->value_list) {
        cond.value_list.push_back(convert_sv_value(val));
      }
    }
    conds.push_back(cond);
  }
}

void Analyze::check_clause(const std::vector<std::string> &tab_names,
                           std::vector<Condition> &conds) {
  // auto all_cols = get_all_cols(tab_names);
  std::vector<ColMeta> all_cols;
  get_all_cols(tab_names, all_cols);
  // Get raw values in where clause
  for (auto &cond : conds) {
    // Infer table name from column name
    cond.lhs_col = check_column(all_cols, cond.lhs_col);
    if (cond.is_rhs_column) {
      cond.rhs_col = check_column(all_cols, cond.rhs_col);
    }
    TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
    auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
    ColType lhs_type = lhs_col->type;
    ColType rhs_type;

    if (cond.is_rhs_val) {
      cond.rhs_val.convert_to_type(lhs_col->type);
      cond.rhs_val.init_raw(lhs_col->len);
      rhs_type = cond.rhs_val.type;
    } else if (cond.is_rhs_column) {
      TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
      auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
      rhs_type = rhs_col->type;
    } else {
      rhs_type = lhs_type;

      // // 检查子查询所涉及的表是否与主查询所涉及表重合 : 不需要检查, 是我看错样例了。
      // for (auto &sub_tab_name : cond.analyzed_sub_query->tables) {
      //   for (auto &tab_name : tab_names) {
      //     if (tab_name == sub_tab_name) {
      //       throw InternalError("sub query tables must unequal to the main query.");
      //     }
      //   }
      // }
    }

    if (!check_type(lhs_type, rhs_type)) {
      throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
    }
  }
}

Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val) {
  Value val;
  if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
    val.set_int(int_lit->val);
  } else if (auto float_lit = std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
    val.set_float(float_lit->val);
  } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
    val.set_str(str_lit->val);
  } else {
    throw InternalError("Unexpected sv value type");
  }
  return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
  std::map<ast::SvCompOp, CompOp> m = {{ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE},
                                       {ast::SV_OP_LT, OP_LT}, {ast::SV_OP_GT, OP_GT},
                                       {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
                                       {ast::SV_OP_IN, OP_IN}, {ast::SV_OP_NOT_IN, OP_NOT_IN}};
  return m.at(op);
}
