/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cassert>
#include <cstddef>
#include <cstring>
#include <memory>
#include <sstream>
#include <string>
#include <vector>

#include "defs.h"
#include "errors.h"
#include "parser/ast.h"
#include "record/rm_defs.h"

enum CompOp { OP_EQ, OP_NE, OP_LT, OP_GT, OP_LE, OP_IN, OP_NOT_IN, OP_GE };

enum CompResult { EQ = 0, LT = -1, GT = 1 };

inline std::string get_aggregation_name(const std::string &str, AggregationType aggregation_type) {
  std::stringstream ret;
  switch (aggregation_type) {
    case AGG_NO: {
      return str;
    } break;
    case AGG_COUNT: {
      ret << "COUNT(" << str << ")";
      return ret.str();
    } break;
    case AGG_COUNT_STAR: {
      ret << "COUNT(*)";
      return ret.str();
    } break;
    case AGG_MAX: {
      ret << "MAX(" << str << ")";
      return ret.str();
    } break;
    case AGG_MIN: {
      ret << "MIN(" << str << ")";
      return ret.str();
    } break;
    case AGG_SUM: {
      ret << "SUM(" << str << ")";
      return ret.str();
    } break;
  }
  return str;
}

inline std::string get_column_name(const std::string &tab_name, const std::string &col_name,
                                   AggregationType aggregation_type) {
  if (tab_name.empty()) {
    return get_aggregation_name(col_name, aggregation_type);
  } else {
    return get_aggregation_name(tab_name + "." + col_name, aggregation_type);
  }
  return "";
}

struct TabCol {
  std::string tab_name;
  std::string col_name;
  std::string alias;  // select的时候起的别名。
  AggregationType aggregation_type{AGG_NO};

  TabCol() = default;

  TabCol(const std::string &tab_name_, const std::string &col_name_, const std::string &alias_,
         AggregationType aggregation_type_)
      : tab_name(tab_name_),
        col_name(col_name_),
        alias(alias_),
        aggregation_type(aggregation_type_) {}

  TabCol(const TabCol &rhs) { *this = rhs; }

  TabCol &operator=(const TabCol &rhs) {
    tab_name = rhs.tab_name;
    col_name = rhs.col_name;
    alias = rhs.alias;
    aggregation_type = rhs.aggregation_type;
    return *this;
  }

  friend bool operator<(const TabCol &x, const TabCol &y) {
    return std::make_pair(x.tab_name, x.col_name) < std::make_pair(y.tab_name, y.col_name);
  }
};

struct Value {
  ColType type;  // type of value
  union {
    int int_val;      // int value
    float float_val;  // float value
  };
  std::string str_val;  // string value

  std::shared_ptr<RmRecord> raw;  // raw record buffer

  void resize(int len) {
    if (type == TYPE_STRING) {
      str_val.resize(len);
    }
  }

  void min(const Value &rhs) {
    switch (type) {
      case TYPE_INT: {
        switch (rhs.type) {
          case TYPE_INT: {
            int_val = std::min(int_val, rhs.int_val);
          } break;
          case TYPE_FLOAT: {
            int_val = std::min(int_val, static_cast<int>(rhs.float_val));
          } break;
          default: {
            throw InternalError("error rhs type max.");
          }
        }
      } break;
      case TYPE_FLOAT: {
        switch (rhs.type) {
          case TYPE_INT: {
            float_val = std::min(float_val, static_cast<float>(rhs.int_val));
          } break;
          case TYPE_FLOAT: {
            float_val = std::min(float_val, rhs.float_val);
          } break;
          default: {
            throw InternalError("error rhs type max.");
          }
        }
      } break;
      case TYPE_STRING: {
        if (rhs.type != TYPE_STRING) {
          throw InternalError("error rhs type max");
        }

        if (compare(str_val, rhs.str_val) == CompResult::GT) {
          str_val = rhs.str_val;
        }
      } break;
      default: {
        throw InternalError("error type max.");
      }
    }
  }

  void max(const Value &rhs) {
    switch (type) {
      case TYPE_INT: {
        switch (rhs.type) {
          case TYPE_INT: {
            int_val = std::max(int_val, rhs.int_val);
          } break;
          case TYPE_FLOAT: {
            int_val = std::max(int_val, static_cast<int>(rhs.float_val));
          } break;
          default: {
            throw InternalError("error rhs type max.");
          }
        }
      } break;
      case TYPE_FLOAT: {
        switch (rhs.type) {
          case TYPE_INT: {
            float_val = std::max(float_val, static_cast<float>(rhs.int_val));
          } break;
          case TYPE_FLOAT: {
            float_val = std::max(float_val, rhs.float_val);
          } break;
          default: {
            throw InternalError("error rhs type max.");
          }
        }
      } break;
      case TYPE_STRING: {
        if (rhs.type != TYPE_STRING) {
          throw InternalError("error rhs type max");
        }

        if (compare(str_val, rhs.str_val) == CompResult::LT) {
          str_val = rhs.str_val;
        }

      } break;
      default: {
        throw InternalError("error type max.");
      }
    }
  }

  void add(const Value &rhs) {
    switch (type) {
      case TYPE_INT: {
        switch (rhs.type) {
          case TYPE_INT: {
            int_val += rhs.int_val;
          } break;
          case TYPE_FLOAT: {
            int_val += static_cast<int>(rhs.float_val);
          } break;
          default: {
            throw InternalError("error rhs type add.");
          }
        }
      } break;
      case TYPE_FLOAT: {
        switch (rhs.type) {
          case TYPE_INT: {
            float_val += static_cast<float>(rhs.int_val);
          } break;
          case TYPE_FLOAT: {
            float_val += rhs.float_val;
          } break;
          default: {
            throw InternalError("error rhs type add.");
          }
        }
      } break;
      default: {
        throw InternalError("error type add.");
      }
    }
  }

  void convert_to_type(ColType convert_type) {
    if (type == convert_type) {
      return;
    }

    switch (convert_type) {
      case TYPE_INT: {
        switch (type) {
          case TYPE_FLOAT: {
            set_int(float_val);
          } break;
          default: {
            throw InternalError("error type convert");
          } break;
        }
      } break;
      case TYPE_FLOAT: {
        switch (type) {
          case TYPE_INT: {
            set_float(int_val);
          } break;
          default: {
            throw InternalError("error type convert");
          } break;
        }
      } break;
      default: {
        throw InternalError("error type convert");
      } break;
    }
  }

  const char *data() const {
    switch (type) {
      case TYPE_INT:
        return reinterpret_cast<const char *>(&int_val);
      case TYPE_FLOAT:
        return reinterpret_cast<const char *>(&float_val);
      case TYPE_STRING:
        return str_val.c_str();
      default: {
        ;
      }
    }
    return nullptr;
  }

  std::string to_string() {
    switch (type) {
      case TYPE_INT:
        return std::to_string(int_val);
      case TYPE_FLOAT:
        return std::to_string(float_val);
      case TYPE_STRING:
        return str_val;
      default: {
        ;
      }
    }
    return std::string();
  }

  CompResult compare(int lhs, int rhs) const {
    return (lhs == rhs ? CompResult::EQ : (lhs < rhs ? CompResult::LT : CompResult::GT));
  }

  CompResult compare(float lhs, float rhs) const {
    return (lhs == rhs ? CompResult::EQ : (lhs < rhs ? CompResult::LT : CompResult::GT));
  }

  // 字符串的比较照搬了MiniOB的方法。
  CompResult compare(const std::string &lhs, const std::string &rhs) const {
    const char *s1 = lhs.c_str();
    const char *s2 = rhs.c_str();
    size_t maxlen = std::min(lhs.size(), rhs.size());
    int result = strncmp(s1, s2, maxlen);

    if (0 != result) {
      return (result < 0) ? CompResult::LT : CompResult::GT;
    }

    if (lhs.size() > maxlen) {
      return (s1[maxlen] - 0) > 0 ? CompResult::GT : CompResult::EQ;
    }

    if (rhs.size() > maxlen) {
      return (0 - s2[maxlen]) < 0 ? CompResult::LT : CompResult::EQ;
    }

    return CompResult::EQ;
  }

  CompResult compare_value(const Value &rhs) const {
    switch (type) {
      case TYPE_INT: {
        if (rhs.type == TYPE_INT) {
          return compare(int_val, rhs.int_val);
        } else if (rhs.type == TYPE_FLOAT) {
          return compare(static_cast<float>(int_val), rhs.float_val);
        } else {
          throw IncompatibleTypeError(coltype2str(type), coltype2str(rhs.type));
        }

        break;
      }
      case TYPE_FLOAT: {
        if (rhs.type == TYPE_INT) {
          return compare(float_val, static_cast<float>(rhs.int_val));
        } else if (rhs.type == TYPE_FLOAT) {
          return compare(float_val, rhs.float_val);
        } else {
          throw IncompatibleTypeError(coltype2str(type), coltype2str(rhs.type));
        }

        break;
      }
      case TYPE_STRING: {
        if (type != rhs.type) {
          throw IncompatibleTypeError(coltype2str(type), coltype2str(rhs.type));
        }
        return compare(str_val, rhs.str_val);

        break;
      }
      default:
        throw InternalError("error type");
    }

    return CompResult::EQ;
  }

  bool check_compare(CompResult res, CompOp comp) const {
    switch (comp) {
      case CompOp::OP_EQ:
        return res == 0;
      case CompOp::OP_GE:
        return res >= 0;
      case CompOp::OP_GT:
        return res > 0;
      case CompOp::OP_LE:
        return res <= 0;
      case CompOp::OP_LT:
        return res < 0;
      case CompOp::OP_NE:
        return res != 0;
      default:
        throw InternalError("error comp op");
    }

    return false;
  }

  void set_int(int int_val_) {
    type = TYPE_INT;
    int_val = int_val_;
  }

  void set_float(float float_val_) {
    type = TYPE_FLOAT;
    float_val = float_val_;
  }

  void set_str(std::string str_val_) {
    type = TYPE_STRING;
    str_val = std::move(str_val_);
  }

  void set_data(const char *data, size_t len = 0) {
    switch (type) {
      case TYPE_INT:
        set_int(*(int *)data);
        break;
      case TYPE_FLOAT:
        set_float(*(float *)data);
        break;
      case TYPE_STRING:
        set_str(std::string(data, len));
        break;
      default:
        throw InternalError("error value type");
    }
  }

  void init_raw(int len) {
    assert(raw == nullptr);
    raw = std::make_shared<RmRecord>(len);
    if (type == TYPE_INT) {
      assert(len == sizeof(int));
      *(int *)(raw->data) = int_val;
    } else if (type == TYPE_FLOAT) {
      assert(len == sizeof(float));
      *(float *)(raw->data) = float_val;
    } else if (type == TYPE_STRING) {
      if (len < (int)str_val.size()) {
        throw StringOverflowError();
      }
      memset(raw->data, 0, len);
      memcpy(raw->data, str_val.c_str(), str_val.size());
    }
  }
};

class Query;
class Plan;
class AbstractExecutor;
struct Condition {
  TabCol lhs_col;   // left-hand side column
  CompOp op;        // comparison operator
  bool is_rhs_val{false};  // true if right-hand side is a value (not a column)
  bool is_rhs_column{false};
  bool is_value_list{false};
  bool is_sub_query{false};
  TabCol rhs_col;   // right-hand side column
  Value rhs_val;    // right-hand side value

  std::shared_ptr<Query> analyzed_sub_query{nullptr};
  std::shared_ptr<Plan> sub_query_plan{nullptr};
  std::shared_ptr<AbstractExecutor> sub_query_executor{nullptr};  // 在begin_tuple中需要打开
  std::vector<Value> value_list;

  Condition();

  Condition(const Condition &rhs);
};

struct SetClause {
  TabCol lhs;
  Value rhs;
};