/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <algorithm>
#include <cstddef>

#include "common/common.h"
#include "defs.h"
#include "errors.h"
#include "execution_defs.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "recovery/log_manager.h"
#include "system/sm.h"
#include "system/sm_meta.h"

class AbstractExecutor {
 public:
  Rid _abstract_rid;

  Context *context_;

  virtual ~AbstractExecutor() = default;

  virtual size_t tupleLen() const { return 0; };

  virtual const std::vector<ColMeta> &cols() const {
    std::vector<ColMeta> *_cols = nullptr;
    return *_cols;
  };

  virtual std::string getType() { return "AbstractExecutor"; };

  virtual void beginTuple(){};

  virtual void nextTuple(){};

  virtual bool is_end() const { return true; };

  virtual Rid &rid() = 0;

  virtual std::unique_ptr<RmRecord> Next() = 0;

  virtual ColMeta get_col_offset(const TabCol &target) { return ColMeta(); };

  std::vector<ColMeta>::const_iterator get_col(const std::vector<ColMeta> &rec_cols,
                                               const TabCol &target) {
    auto pos = std::find_if(rec_cols.begin(), rec_cols.end(), [&](const ColMeta &col) {
      return col.tab_name == target.tab_name && col.name == target.col_name;
    });
    if (pos == rec_cols.end()) {
      throw ColumnNotFoundError(target.tab_name + '.' + target.col_name);
    }
    return pos;
  }

  Value get_value(const RmRecord *rcd, const ColMeta &col) const {
    char *data = rcd->data + col.offset;
    Value val;
    val.type = col.type;
    val.set_data(data, col.len);
    // val.init_raw(col.len);
    return val;
  }

  void open_conditions(std::vector<Condition> &conditions) {
    for (auto &cond : conditions) {
      if (cond.is_sub_query) {
        if (cond.sub_query_executor->cols().size() != 1) {
          throw InternalError("simple sub query result should be one row.");
        }

        // 对单行单列的进行处理。
        if (cond.op != OP_IN && cond.op != OP_NOT_IN) {
          cond.sub_query_executor->beginTuple();

          if (cond.sub_query_executor->is_end()) {
            throw InternalError("simple sub query result is empty.");
          }

          auto rec = cond.sub_query_executor->Next();

          cond.sub_query_executor->nextTuple();

          if (!cond.sub_query_executor->is_end()) {
            throw InternalError("simple sub query result should be one line.");
          }

          cond.is_rhs_val = true;
          cond.is_sub_query = false;
          auto &col_meta = cond.sub_query_executor->cols().front();
          cond.rhs_val.type = col_meta.type;
          cond.rhs_val.set_data(rec->data + col_meta.offset, col_meta.len);
        }
      } else if (cond.is_value_list) {
        if (cond.op != OP_IN && cond.op != OP_NOT_IN) {
          if (cond.value_list.size() != 1) {
            throw InternalError("value list size shoule be 1.");
          }

          cond.is_rhs_val = true;
          cond.is_value_list = false;
          cond.rhs_val = cond.value_list.front();
        }
      }
    }
  }

  // 传入参数，左，右元组。条件，两元组ColMeta。
  bool check_tuple(RmRecord *left, RmRecord *right, const Condition &cond,
                   const std::vector<ColMeta> &left_cols, const std::vector<ColMeta> &right_cols) {
    auto &lhs_col = cond.lhs_col;
    auto &lhs_col_meta = *get_col(left_cols, lhs_col);
    auto lhs_val = get_value(left, lhs_col_meta);

    // 这句话的意思是： 如果 !is_rhs_val 并且 rhs_col_iter == end(),说明是子查询，rhs_col_meta =
    // 子查询cols()的front().
    //                如果 is_rhs_val,
    //                那么这个时候rhs_col_meta是用不到的，为了保持三目运算符一致，返回lhs_col_meta。
    //                如果 !is_rhs_val 并且 rhs_col_iter!=end()，那么说明是两列查询，。
    // 这么麻烦使用三目运算符的原因是，想用引用.....
    ColMeta rhs_col_meta;

    if (cond.is_rhs_val || cond.is_value_list) {
      ;
    } else if (cond.is_sub_query) {
      rhs_col_meta = cond.sub_query_executor->cols().front();
    } else if (cond.is_rhs_column) {
      rhs_col_meta = *get_col(right_cols, cond.rhs_col);
    } else {
      throw InternalError("error condition.");
    }

    if (cond.is_rhs_val || cond.is_rhs_column) {
      auto rhs_val = (cond.is_rhs_val ? cond.rhs_val : get_value(right, rhs_col_meta));
      return lhs_val.check_compare(lhs_val.compare_value(rhs_val), cond.op);
    }

    if (cond.is_value_list) {
      if (cond.op != OP_IN && cond.op != OP_NOT_IN) {
        throw InternalError("error condition.");
      }

      for (auto &val : cond.value_list) {
        bool res = (lhs_val.compare_value(val) == CompResult::EQ);

        if (res) {
          return res && cond.op == OP_IN;
        }
      }

      return cond.op != OP_IN;
    }

    if (cond.is_sub_query) {
      if (cond.op != OP_IN && cond.op != OP_NOT_IN) {
        throw InternalError("error condition.");
      }

      Value rhs_val;

      for (cond.sub_query_executor->beginTuple(); !cond.sub_query_executor->is_end();
           cond.sub_query_executor->nextTuple()) {
        auto row = cond.sub_query_executor->Next();
        rhs_val.type = rhs_col_meta.type;

        // 其实不用 + offset，因为单行单列，所以offset==0.
        rhs_val.set_data(row->data + rhs_col_meta.offset, rhs_col_meta.len);

        bool res = (lhs_val.compare_value(rhs_val) == CompResult::EQ);

        if (res) {
          return (res && cond.op == OP_IN);
        }
      }

      return (cond.op != OP_IN);
    }

    throw InternalError("error condition.");
    return false;
  }

  bool check_conditions(RmRecord *left, RmRecord *right, const std::vector<Condition> &cond,
                        const std::vector<ColMeta> &left_cols,
                        const std::vector<ColMeta> &right_cols) {
    for (auto &conditon : cond) {
      if (!check_tuple(left, right, conditon, left_cols, right_cols)) {
        return false;
      }
    }

    return true;
  }
};