/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>
#include <memory>

#include "defs.h"
#include "errors.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "index/ix_index_handle.h"
#include "record/rm_defs.h"
#include "system/sm.h"
#include "system/sm_meta.h"

class UpdateExecutor : public AbstractExecutor {
 private:
  TabMeta tab_;
  std::vector<Condition> conds_;
  RmFileHandle *fh_;
  std::vector<Rid> rids_;
  std::string tab_name_;
  std::vector<SetClause> set_clauses_;
  SmManager *sm_manager_;

 public:
  UpdateExecutor(SmManager *sm_manager, const std::string &tab_name,
                 std::vector<SetClause> set_clauses, std::vector<Condition> conds,
                 std::vector<Rid> rids, Context *context) {
    sm_manager_ = sm_manager;
    tab_name_ = tab_name;
    set_clauses_ = set_clauses;
    tab_ = sm_manager_->db_.get_table(tab_name);
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    conds_ = conds;
    rids_ = rids;
    context_ = context;
  }

  // 当更新的值违反了唯一性索引，需要抛出异常。
  // 更新麻烦的点在于，如果被更新的tuple不止一个，那么在恢复的时候很难办，所以不得不保存旧的tuples。
  std::unique_ptr<RmRecord> Next() override {
    size_t rid_i;

    std::vector<RmRecord> old_recs;  // 保存旧的tuples。

    for (rid_i = 0; rid_i < rids_.size(); ++rid_i) {
      auto &rid = rids_[rid_i];
      // 新份tuple，更新之后进行修改。
      auto new_rec = fh_->get_record(rid, context_);
      auto old_rec = std::make_unique<RmRecord>(*new_rec);

      // 旧份tuple，不进行更新，后面索引操作有用。

      for (auto &set_clause : set_clauses_) {
        auto col_meta_iter = std::find_if(
            tab_.cols.begin(), tab_.cols.end(),
            [&set_clause](ColMeta &col) { return col.name == set_clause.lhs.col_name; });

        if (col_meta_iter == tab_.cols.end()) {
          throw InternalError("The field not exists in the table.");
        }

        set_clause.rhs.convert_to_type(col_meta_iter->type);
        set_clause.rhs.resize(col_meta_iter->len);

        memcpy(new_rec->data + col_meta_iter->offset, set_clause.rhs.data(), col_meta_iter->len);
      }

      // 用新tuple更新。
      fh_->update_record(rid, new_rec->data, context_);

      // 处理索引 ：
      size_t i;
      for (i = 0; i < tab_.indexes.size(); ++i) {
        auto &index_meta = tab_.indexes[i];
        auto index =
            sm_manager_->ihs_
                .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols))
                .get();

        RmRecord update_rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(new_rec.get(), &update_rec);

        // 修改之前的tuple生成对应索引的index_tuple
        RmRecord rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(old_rec.get(), &rec);

        // 如果更新之后的值和更新之前的值一样，那么不需要进行更新。不要返回failure
        // 这个地方是想到哪里试试，发现更新重复的时候会返回failure，我本来觉得没啥的
        // 但是以防万一打开了mysql对照一下，发现mysql的处理是不报错，所以对照mysql的行为写了。
        // 这里是题目三最后一个bug，改完之后就过了，记录一下。
        if (0 == ix_compare(update_rec.data, rec.data, index->get_index_col_types(),
                            index->get_index_col_lens())) {
          continue;
        }

        std::vector<Rid> rids;
        index->get_value(update_rec.data, &rids, context_->txn_);

        if (!rids.empty()) {  // 新值已经存在，违反唯一性原则，需要恢复并抛出异常。
          break;
        }

        index->get_value(rec.data, &rids, context_->txn_);

        // 下面这两个应该是不可能出现的情况，如果出现了那么说明该debug了，而不是恢复现场。
        if (rids.empty()) {  // 找到了。
          throw InternalError("rid not exists");
        } else if (rids.front() != rid) {
          throw InternalError("rid not equal");
        }

        index->delete_entry(rec.data, context_->txn_);
        index->insert_entry(update_rec.data, rid, context_->txn_);
      }

      if (i == tab_.indexes.size()) {
        old_recs.push_back(*old_rec);  // 这个过。
        continue;
      }

      // 不相等，恢复现场。
      fh_->update_record(rid, old_rec->data, context_);

      for (size_t j = 0; j < i; ++j) {  // 第i个索引没有更新，
        auto &index_meta = tab_.indexes[j];
        auto index =
            sm_manager_->ihs_
                .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols))
                .get();

        RmRecord update_rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(new_rec.get(), &update_rec);

        // 修改之前的tuple生成对应索引的index_tuple
        RmRecord rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(old_rec.get(), &rec);

        // 删除更新值，插入老值。
        index->delete_entry(update_rec.data, context_->txn_);
        index->insert_entry(rec.data, rid, context_->txn_);
      }

      break;
    }

    // 所有的tuple都被正常的更新了。
    if (rid_i == rids_.size()) {
      return nullptr;
    }

    // 存在违反索引唯一性的情况出现。相当于undo了吧。
    for (size_t j = 0; j < rid_i; ++j) {
      auto &rid = rids_[j];
      auto &old_rec = old_recs[j];
      auto new_rec = std::make_unique<RmRecord>(old_rec);

      for (auto &set_clause : set_clauses_) {
        auto col_meta_iter = std::find_if(
            tab_.cols.begin(), tab_.cols.end(),
            [&set_clause](ColMeta &col) { return col.name == set_clause.lhs.col_name; });

        if (col_meta_iter == tab_.cols.end()) {
          throw InternalError("The field not exists in the table.");
        }

        set_clause.rhs.convert_to_type(col_meta_iter->type);
        set_clause.rhs.resize(col_meta_iter->len);
        
        memcpy(new_rec->data + col_meta_iter->offset, set_clause.rhs.data(), col_meta_iter->len);
      }

      fh_->update_record(rid, old_rec.data, context_);  // 表恢复成原来的tuple。

      for (auto &index_meta : tab_.indexes) {
        auto index =
            sm_manager_->ihs_
                .at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols))
                .get();

        RmRecord update_rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(new_rec.get(), &update_rec);

        // 修改之前的tuple生成对应索引的index_tuple
        RmRecord rec(index_meta.col_tot_len);
        index_meta.make_index_tuple(&old_rec, &rec);

        // 删除更新值，插入老值。
        index->delete_entry(update_rec.data, context_->txn_);
        index->insert_entry(rec.data, rid, context_->txn_);
      }
    }

    throw InternalError("updated value has exists in index.");
    return nullptr;
  }

  Rid &rid() override { return _abstract_rid; }
  bool is_end() const override { return true; }
};